import greenfoot.Actor;
import greenfoot.Greenfoot;
import greenfoot.GreenfootImage;
import greenfoot.World;

import java.awt.Color;

/**
 * A   <a href="http://en.wikipedia.org/wiki/Conway's_Game_of_Life" target="_blank">game of live</a> world consisting of lines and cells.
 * 
 */
public class LifeWorld extends World {

   static private final int
     initialLivingPercentage = 25
     ,width = 80
     ,height = 50;
   
   public static final int cellSize = 10; // Minimum of 3 pixel in width here yields a (cellSize-2)x(cellSize-2) pixel visible cell
   
   static public GreenfootImage 
     imageAlive = new GreenfootImage(cellSize, cellSize)
     ,imageDead = new GreenfootImage(cellSize, cellSize);

   // "Class" constructor initializing static fields. You may want to read Ivor Horton's
   // "Using Initialization Blocks" / "Initializing Data Members" subsection.
   static {              
      imageDead.setColor(Color.WHITE);
      imageDead.fillRect(1, 1, cellSize - 1, cellSize - 1);
      
      imageAlive.setColor(Color.black);
      imageAlive.fillRect(1, 1, cellSize - 1, cellSize - 1);
    }

   private final GreenfootImage gridImage;
   private final Cell[][] cells;

   /**
    * Constructor for objects of class Grid. Calls methods to draw the grid
    * lines, to create the array of randomly initialized cells .
    * 
    */
   public LifeWorld() {
      super(width, height, cellSize);
      gridImage = getBackground();
      drawGrid();
      cells = new Cell[width][height];  // Create cells array container.
      createCells();

      // Having created all cells we may now determine and set each cell's set of neighbors.
      for (int x = 0; x < width; x++) {  
         for (int y = 0; y < height; y++) {
            cells[x][y].setNeighbors(getCellNeighbours(x, y));
         }
      }   
   }

   /**
    * Tell each cell to save its current state in preparation of the game's
    * next step. According to {@link World#act()} this happens prior to
    * calling any actor's {@link Actor#act()} methods methods.
    */
   public void act() {
      for (int x = 0; x < width; x++) {
         for (int y = 0; y < height; y++) {
            cells[x][y].saveCurrentToPreviousState();
         }
      }
   }

   /**
    * Draw the cell separating grid lines.
    */
   private void drawGrid() {
      gridImage.setColor(Color.red);

      final int endX = cellSize * width,
            endY = cellSize * height;

      for (int x = 0; x < endX; x += cellSize) { // Drawing vertical grid lines. Diagram
         gridImage.drawLine(x, 0, x, endY);       // is being scaled by cellSize:
      }                                          // 
      //            0   1   2 ...(width - 1)
      //         0  |   |   | ...  |
      //         1  |   |   | ...  |
      //         2  |   |   | ...  |
      //           ... ... ...    ...   
      // (height-1) |   |   | ...  |

      for (int y = 0; y < endY; y += cellSize) { // Drawing horizontal grid lines. Diagram 
         gridImage.drawLine(0, y, endX, y);       // is being scaled by cellSize:
      }                                          //
      //            0   1   2 ...(width - 1)
      //         0  __________..._____
      //         1  __________..._____ 
      //         2  __________..._____
      //           .....................
      // (height-1) __________..._____
   }

   /** Create the array of cells and randomly turn some of them on
    *  according to {@link #setInitialLivingPercentage(int)}.
    */
   private void createCells() {
      for (int x = 0; x < width; x++) {
         for (int y = 0; y < height; y++) {
            // New cell will have a probability of 
            // LifeWorld.initialLivingPercentage / 100 to be created
            // in "live" state.
            final Cell c = new Cell(getRandomLifeOrDeadState());
            
            cells[x][y] = c;    // A separate array allows for faster access than via
                                // Greenfoot's own "world" object accounting mechanism.
            addObject(c, x, y); // Add to LifeWorld as well.
         }
      }
   }

   /**
    * Find the set of all neighbors (n) of a given cell (c). A cell may have
    * a maximum of eight , five or just three neighbors depending on its
    * relative position to the world's boundary:
    * 
    *  Middle:    Edge:        Corner:
    *  8 Neigh-   5 Neigh-     3 Neigh-
    *    bours      bours        bours
    * .........   _________     _____
    * ...nnn...   ...ncn...    |cn...   3 other corner positions left / right / bottom
    * ...ncn...   ...nnn...    |nn...   and 3 more edge positions upper right /
    * ...nnn...   .........    |.....   lower left / lower right omitted). So we have
    * .........   .........    |.....   9 different world boundary related situations. 
    * 
    * @param x The given cell's x-position
    * @param y The given cell's y-position
    * @return The array of all neighboring cells.
    */
   private Cell[] getCellNeighbours(final int x, final int y) {
      // Rectangular limits corresponding to the 9 different boundary
      // related situations being described in the above Javadoc.
      //
      final int
      xMin = Math.max(              0, x - 1),       // Possibly limited by left border
      xMax = Math.min( getWidth() - 1, x + 1),       // Possibly limited by right border

      yMin = Math.max(0              , y - 1),       // Possibly limited by top border
      yMax = Math.min(getHeight() - 1, y + 1);       // Possibly limited by bottom border

      // Allocate array of appropriate size
      // to hold references to neighbouring cells
      final Cell[] neighbours = new Cell[             // Middle     / Edge      / Corner:
         (xMax - xMin + 1) * (yMax - yMin + 1) - 1]; // 3*3-1 == 5 / 2*3-1 == 5/ 2*2-1 == 3 neighbours
      
      // Fill in cell neighbours
      int neighborIndexCount = 0;
      for (int dx = xMin; dx <= xMax; dx++) {
         for (int dy = yMin; dy <= yMax; dy++) {
            if (dx != x || dy != y) {// Exclude center cell, aka "this"
               neighbours[neighborIndexCount++] = cells[dx][dy];
            }
         }
      }
      return neighbours;
   }

   /** Randomly return the two states {@link CellState#ALIVE} and {@link CellState#DEAD}. The
    *  corresponding probabilities are {@link #setInitialLivingPercentage(int)} / 100 and
    *  its complement.
    * 
    * @return Either {@link CellState#ALIVE} or {@link CellState#DEAD}.
    */
   public static CellState getRandomLifeOrDeadState() {
      return Greenfoot.getRandomNumber(100) < initialLivingPercentage ? 
            CellState.ALIVE : CellState.DEAD;
   }
}
